package spaces

import "time"

type State int

const (
	StateUnknown     State = iota
	StateNew               // Новое пространство, БД еще не инициализирована
	StateReady             // Пространство готово к использованию
	StatePreparing         // Пространство готовится в использованию
	StateMaintenance       // Пространство находится на обслуживании

	// StateMigrationScheduled Для пространства запланирована миграция
	// При планировании запуска миграции пространства это состояние выступает в роли Lock'а, который не
	// позволяет запустить несколько миграций одновременно. Логика планировщика миграций сначала пытается
	// установить для пространства состояние StateMigrationScheduled, что удается только в двух случаях:
	// пространство находится в состоянии StateReady или же в состояниях StateMigration, StateMigrationScheduled,
	// но последнее обновление состояния было более 15-ти минут назад
	StateMigrationScheduled
	StateMigration // Пространство в состоянии выполнения миграций
	StateDeleting  // Пространство удаляется, данные будут удалены
	StateError     // Произошла ошибка при конфигурации пространства (см. Space.Error)
)

func (s State) String() string {
	switch s {
	case StateNew:
		return "New"
	case StateReady:
		return "Ready"
	case StatePreparing:
		return "Preparing"
	case StateMaintenance:
		return "Maintenance"
	case StateMigrationScheduled:
		return "MigrationScheduled"
	case StateMigration:
		return "Migration"
	case StateDeleting:
		return "Deleting"
	case StateError:
		return "Error"
	default:
		return "Unknown"
	}
}

var (
	ReadAllowedStates = []State{
		StateNew,
		StateReady,
		StateMigrationScheduled,
		StateMigration,
	}

	WriteAllowedStates = []State{
		StateNew,
		StateReady,
	}
)

type Config struct {
	Features []string // Deprecated Возможности используемые пространством
}

const StateInfoEmpty = "EMPTY"

type Space struct {
	ID          string `json:"id,omitempty" bson:"_id"`
	OrgID       string `json:"org_id,omitempty" bson:"org_id"`
	Name        string `json:"name,omitempty" bson:"name"`
	Description string `json:"desc,omitempty" bson:"desc"`
	// TransferToOrg - идентификатор организации, в которую был запрошен перенос пространства
	TransferToOrg string `json:"transfer_to_org" bson:"transfer_to_org"`

	Config    *Config    `json:"config" bson:"config"`
	StateInfo *StateInfo `json:"state_info,omitempty" bson:"state_info,omitempty"`
}

type StateInfo struct {
	State     State     `json:"state" bson:"state,omitempty"`
	Info      string    `json:"info" bson:"info,omitempty"`
	Time      time.Time `json:"time,omitempty" bson:"time,omitempty"`
	DBVersion uint32    `json:"db_version" bson:"db_version,omitempty"`
}

func (s Space) Clone() *Space {
	return &s
}
