package zap

import (
	"slices"

	"go.uber.org/zap"
	"go.uber.org/zap/zapcore"
)

type FilterFunc func(entry zapcore.Entry, fields []zapcore.Field) bool

func ContainsField(field zapcore.Field) FilterFunc {
	return func(_ zapcore.Entry, fields []zapcore.Field) bool {
		for _, f := range fields {
			if f.Equals(field) {
				return true
			}
		}
		return false
	}
}

func ContainsKey(key string) FilterFunc {
	return func(entry zapcore.Entry, fields []zapcore.Field) bool {
		for _, f := range fields {
			if f.Key == key {
				return true
			}
		}
		return false
	}
}

func Or(filters ...FilterFunc) FilterFunc {
	return func(entry zapcore.Entry, fields []zapcore.Field) bool {
		for _, f := range filters {
			if f(entry, fields) {
				return true
			}
		}
		return false
	}
}

func Not(filter FilterFunc) FilterFunc {
	return func(entry zapcore.Entry, fields []zapcore.Field) bool {
		return !filter(entry, fields)
	}
}

type filterCore struct {
	zapcore.Core

	filters []FilterFunc

	// context хранит контекст записей ядра, передаваемых при вызове With.
	// В методе Write передаются только поля конкретной записи, но мы также хотим учитывать поля контекста ядра.
	context []zap.Field
}

// WithFilters - добавить фильтры, которые будут применяться при записи лога (вызове `core.Write`)
// Метод `core.Write` будет вызван только в случае, когда результат всех фильтров `true`
//
// Обратить внимание, фильтр не применяется к полям, которые были добавлены в `core` через вызов `core.With`
// до вызова WithFilters. Пример:
//
//	l, _ := zap.NewDevelopment()
//	core := l.Core().With([]zapcore.Field{zap.Int("a", 5)})
//	core = WithFilters(core, ContainsField(zap.Int("a", 5)))
//
//	logger := zap.New(core)
//	logger.Info("Test log") // НЕ будет записан
//	logger.Info("Test log", zap.Int("a", 5)) // будет записан
func WithFilters(core zapcore.Core, filters ...FilterFunc) zapcore.Core {
	return &filterCore{
		Core:    core,
		filters: filters,
	}
}

func (core *filterCore) With(fields []zapcore.Field) zapcore.Core {
	return &filterCore{
		Core:    core.Core.With(fields),
		filters: core.filters,
		context: slices.Concat(core.context, fields),
	}
}

func (core *filterCore) Check(entry zapcore.Entry, checkedEntry *zapcore.CheckedEntry) *zapcore.CheckedEntry {
	if core.Core.Enabled(entry.Level) {
		return checkedEntry.AddCore(entry, core)
	}
	return checkedEntry
}

func (core *filterCore) Write(entry zapcore.Entry, fields []zapcore.Field) error {
	all := slices.Concat(core.context, fields)
	for _, filter := range core.filters {
		if !filter(entry, all) {
			return nil
		}
	}
	return core.Core.Write(entry, fields)
}
